﻿#include "pch.h"
#include "GeometryVarietiesRenderer.h"

using namespace GeometryExperimentation;

using namespace D2D1;
using namespace DirectX;
using namespace Microsoft::WRL;
using namespace Platform;
using namespace Windows::Foundation;
using namespace Windows::Foundation::Collections;
using namespace Windows::UI::Core;

GeometryVarietiesRenderer::GeometryVarietiesRenderer() : m_fillMode(D2D1_FILL_MODE_ALTERNATE),
                                                         m_renderingOption(RenderingOption::None)
{
}

void GeometryVarietiesRenderer::CreateDeviceIndependentResources()
{
    DirectXBase::CreateDeviceIndependentResources();

    // Create square-wave geometry
    HRESULT hr = m_d2dFactory->CreatePathGeometry(&m_squareWaveGeometry);

    if (FAILED(hr))
        throw ref new COMException(hr, "factory->CreatePathGeometry failure");

    ComPtr<ID2D1GeometrySink> geometrySink;
    hr = m_squareWaveGeometry->Open(&geometrySink);

    if (FAILED(hr))
        throw ref new COMException(hr, "pathGeometry->Open failure");

    geometrySink->BeginFigure(Point2F(-250, 50), D2D1_FIGURE_BEGIN_HOLLOW);
    geometrySink->AddLine(Point2F(-250, -50));
    geometrySink->AddLine(Point2F(-150, -50));
    geometrySink->AddLine(Point2F(-150,  50));
    geometrySink->AddLine(Point2F(-50,   50));
    geometrySink->AddLine(Point2F(-50,  -50));
    geometrySink->AddLine(Point2F( 50,  -50));
    geometrySink->AddLine(Point2F( 50,   50));
    geometrySink->AddLine(Point2F(150,   50));
    geometrySink->AddLine(Point2F(150,  -50));
    geometrySink->AddLine(Point2F(250,  -50));
    geometrySink->AddLine(Point2F(250,   50));
    geometrySink->EndFigure(D2D1_FIGURE_END_OPEN);

    hr = geometrySink->Close();

    if (FAILED(hr))
        throw ref new COMException(hr, "geometrySink->Close failure");

    // Create star geometry and translate it
    ComPtr<ID2D1PathGeometry> starGeometry;
    hr = CreateFivePointedStar(150, &starGeometry);
    
    if (FAILED(hr))
        throw ref new COMException(hr, "CreateFivePointedStar failure");

    hr = m_d2dFactory->CreateTransformedGeometry(starGeometry.Get(), 
                                                 Matrix3x2F::Translation(0, -200), 
                                                 &m_starGeometry);
    
    if (FAILED(hr))
        throw ref new COMException(hr, "factory->CreateTransformedGeometry failure");

    // Create infinity geometry and translate it
    ComPtr<ID2D1PathGeometry> infinityGeometry;
    hr = CreateInfinitySign(100, &infinityGeometry);
    
    if (FAILED(hr))
        throw ref new COMException(hr, "CreateInfinitySign failure");

    hr = m_d2dFactory->CreateTransformedGeometry(infinityGeometry.Get(), 
                                                 Matrix3x2F::Translation(0, 200), 
                                                 &m_infinityGeometry);
    if (FAILED(hr))
        throw ref new COMException(hr, "factory->CreateTransformedGeometry failure");

    // Create geometry group
    CreateGeometryGroup();

    // Create stroke styles
    hr = m_d2dFactory->CreateStrokeStyle(StrokeStyleProperties(D2D1_CAP_STYLE_ROUND, 
                                                               D2D1_CAP_STYLE_ROUND, 
                                                               D2D1_CAP_STYLE_ROUND, 
                                                               D2D1_LINE_JOIN_ROUND, 
                                                              10.0f, D2D1_DASH_STYLE_SOLID, 0), 
                                        nullptr, 0, &m_roundedStrokeStyle);
    if (FAILED(hr))
        throw ref new COMException(hr, "factory->CreateStrokeStyle failure");

    hr = m_d2dFactory->CreateStrokeStyle(StrokeStyleProperties(D2D1_CAP_STYLE_ROUND, 
                                                               D2D1_CAP_STYLE_ROUND, 
                                                               D2D1_CAP_STYLE_ROUND, 
                                                               D2D1_LINE_JOIN_ROUND, 
                                                              10.0f, D2D1_DASH_STYLE_DOT, 0), 
                                        nullptr, 0, &m_dottedStrokeStyle);
    if (FAILED(hr))
        throw ref new COMException(hr, "factory->CreateStrokeStyle failure");
}

void GeometryVarietiesRenderer::CreateDeviceResources()
{
	DirectXBase::CreateDeviceResources();

    HRESULT hr = m_d2dContext->CreateSolidColorBrush(ColorF(ColorF::Black), &m_blackBrush);

    if (FAILED(hr))
        throw ref new COMException(hr, "context->CreateSolidColorBrush failure");

    hr = m_d2dContext->CreateSolidColorBrush(ColorF(ColorF::Red), &m_redBrush);

    if (FAILED(hr))
        throw ref new COMException(hr, "context->CreateSolidColorBrush failure");
}

void GeometryVarietiesRenderer::CreateWindowSizeDependentResources()
{
	DirectXBase::CreateWindowSizeDependentResources();
}

HRESULT GeometryVarietiesRenderer::CreateInfinitySign(float radius, ID2D1PathGeometry** ppPathGeometry)
{
	if (ppPathGeometry == nullptr)
		return E_POINTER;

    HRESULT hr = m_d2dFactory->CreatePathGeometry(ppPathGeometry);
    ComPtr<ID2D1GeometrySink> geometrySink;

    if (SUCCEEDED(hr))
    {
        hr = (*ppPathGeometry)->Open(&geometrySink);
    }

    if (SUCCEEDED(hr))
    {
        float x1 = 0.50f * radius;
        float x2 = 0.95f * radius;
        float x3 = 1.50f * radius;
        float x4 = 2.05f * radius;
        float x5 = 2.50f * radius;

        float y1 = 0.50f * radius;
        float y2 = 0.55f * radius;
        float y3 = 1.00f * radius;

        geometrySink->BeginFigure(Point2F(-x3, -y3), D2D1_FIGURE_BEGIN_FILLED);
        geometrySink->AddBezier(BezierSegment(Point2F(-x4, -y3), Point2F(-x5, -y2), Point2F(-x5, 0)));
        geometrySink->AddBezier(BezierSegment(Point2F(-x5, y2), Point2F(-x4, y3), Point2F(-x3, y3)));
        geometrySink->AddBezier(BezierSegment(Point2F(-x2, y3), Point2F(-x1, y1), Point2F(0, 0)));
        geometrySink->AddBezier(BezierSegment(Point2F(x1, -y1), Point2F(x2, -y3), Point2F(x3, -y3)));
        geometrySink->AddBezier(BezierSegment(Point2F(x4, -y3), Point2F(x5, -y2), Point2F(x5, 0)));
        geometrySink->AddBezier(BezierSegment(Point2F(x5, y2), Point2F(x4, y3), Point2F(x3, y3)));
        geometrySink->AddBezier(BezierSegment(Point2F(x2, y3), Point2F(x1, y1), Point2F(0, 0)));
        geometrySink->AddBezier(BezierSegment(Point2F(-x1, -y1), Point2F(-x2, -y3), Point2F(-x3, -y3)));
        geometrySink->EndFigure(D2D1_FIGURE_END_CLOSED);

        hr = geometrySink->Close();
    }

    return hr;
}

HRESULT GeometryVarietiesRenderer::CreateFivePointedStar(float radius, ID2D1PathGeometry** ppPathGeometry)
{
	if (ppPathGeometry == nullptr)
		return E_POINTER;

    HRESULT hr = m_d2dFactory->CreatePathGeometry(ppPathGeometry);
    ComPtr<ID2D1GeometrySink> geometrySink;

    if (SUCCEEDED(hr))
    {
        hr = (*ppPathGeometry)->Open(&geometrySink);
    }

    if (SUCCEEDED(hr))
    {
        geometrySink->BeginFigure(Point2F(0, -radius), D2D1_FIGURE_BEGIN_FILLED);

        for (float angle = 2 * XM_2PI / 5; angle < 2 * XM_2PI; angle += 2 * XM_2PI / 5)
        {
            float sin, cos;
            D2D1SinCos(angle, &sin, &cos);
            geometrySink->AddLine(Point2F(radius * sin, -radius * cos));
        }

        geometrySink->EndFigure(D2D1_FIGURE_END_CLOSED);
        hr = geometrySink->Close();
    }

    return hr;
}

void GeometryVarietiesRenderer::CreateGeometryGroup()
{
    ID2D1Geometry* geometries[] = 
    {
        m_squareWaveGeometry.Get(),
        m_starGeometry.Get(),
        m_infinityGeometry.Get()
    };

    HRESULT hr = m_d2dFactory->CreateGeometryGroup(m_fillMode, geometries, ARRAYSIZE(geometries), &m_geometryGroup);

    if (FAILED(hr))
        throw ref new COMException(hr, "factory->CreateGeometryGroup failure");

    // Create simplified geometry
    hr = m_d2dFactory->CreatePathGeometry(&m_simplifiedGeometry);

    if (FAILED(hr))
        throw ref new COMException(hr, "factory->CreatePathGeometry failure");

    ComPtr<ID2D1GeometrySink> geometrySink;
    hr = m_simplifiedGeometry->Open(&geometrySink);

    if (FAILED(hr))
        throw ref new COMException(hr, "pathGeometry->Open failure");

    hr = m_geometryGroup->Simplify(D2D1_GEOMETRY_SIMPLIFICATION_OPTION_CUBICS_AND_LINES, IdentityMatrix(), geometrySink.Get());

    if (FAILED(hr))
        throw ref new COMException(hr, "geometry->Simpify failure");

    hr = geometrySink->Close();

    if (FAILED(hr))
        throw ref new COMException(hr, "geometrySink->Close failure");

    // Create grossly simplified geometry
    hr = m_d2dFactory->CreatePathGeometry(&m_grosslySimplifiedGeometry);

    if (FAILED(hr))
        throw ref new COMException(hr, "factory->CreatePathGeometry failure");

    hr = m_grosslySimplifiedGeometry->Open(&geometrySink);

    if (FAILED(hr))
        throw ref new COMException(hr, "pathGeometry->Open failure");

    hr = m_geometryGroup->Simplify(D2D1_GEOMETRY_SIMPLIFICATION_OPTION_LINES, IdentityMatrix(), 20, geometrySink.Get());

    if (FAILED(hr))
        throw ref new COMException(hr, "geometry->Simpify failure");

    hr = geometrySink->Close();

    if (FAILED(hr))
        throw ref new COMException(hr, "geometrySink->Close failure");

    // Create outlined geometry
    hr = m_d2dFactory->CreatePathGeometry(&m_outlinedGeometry);

    if (FAILED(hr))
        throw ref new COMException(hr, "factory->CreatePathGeometry failure");

    hr = m_outlinedGeometry->Open(&geometrySink);

    if (FAILED(hr))
        throw ref new COMException(hr, "pathGeometry->Open failure");

    hr = m_geometryGroup->Outline(IdentityMatrix(), geometrySink.Get());
    
    if (FAILED(hr))
        throw ref new COMException(hr, "geometry->Outline failure");

    hr = geometrySink->Close();

    if (FAILED(hr))
        throw ref new COMException(hr, "geometrySink->Close failure");

    // Create widened geometry
    hr = m_d2dFactory->CreatePathGeometry(&m_widenedGeometry);

    if (FAILED(hr))
        throw ref new COMException(hr, "factory->CreatePathGeometry failure");

    hr = m_widenedGeometry->Open(&geometrySink);

    if (FAILED(hr))
        throw ref new COMException(hr, "pathGeometry->Open failure");

    hr = m_geometryGroup->Widen(20, m_roundedStrokeStyle.Get(), IdentityMatrix(), geometrySink.Get());

    if (FAILED(hr))
        throw ref new COMException(hr, "geometry->Widen failure");

    hr = geometrySink->Close();

    if (FAILED(hr))
        throw ref new COMException(hr, "geometrySink->Close failure");

    // Create outlined widened geometry
    hr = m_d2dFactory->CreatePathGeometry(&m_outlinedWidenedGeometry);

    if (FAILED(hr))
        throw ref new COMException(hr, "factory->CreatePathGeometry failure");

    hr = m_outlinedWidenedGeometry->Open(&geometrySink);
    if (FAILED(hr))
        throw ref new COMException(hr, "pathGeometry->Open failure");

    hr = m_widenedGeometry->Outline(IdentityMatrix(), geometrySink.Get());

    if (FAILED(hr))
        throw ref new COMException(hr, "geometry->Outline failure");

    hr = geometrySink->Close();

    if (FAILED(hr))
        throw ref new COMException(hr, "geometrySink->Close failure");
}

void GeometryVarietiesRenderer::Update(float timeTotal, float timeDelta)
{
    // Recreate m_animatedStrokeStyle
    if (m_renderingOption == RenderingOption::AnimatedDotOffset)
    {
        float dotOffset = fmodf(2 * timeTotal, 2);

        m_d2dFactory->CreateStrokeStyle(StrokeStyleProperties(D2D1_CAP_STYLE_ROUND, 
                                                          D2D1_CAP_STYLE_ROUND, 
                                                          D2D1_CAP_STYLE_ROUND, 
                                                          D2D1_LINE_JOIN_ROUND, 
                                                          10.0f, D2D1_DASH_STYLE_DOT, 
                                                          dotOffset), 
                                    nullptr, 0, &m_animatedStrokeStyle);
    }
}

void GeometryVarietiesRenderer::Render()
{
	m_d2dContext->BeginDraw();
    m_d2dContext->Clear(ColorF(ColorF::White));

	// Put origin in center of screen
	Matrix3x2F translation = Matrix3x2F::Translation(m_windowBounds.Width / 2.0f,
		                                             m_windowBounds.Height / 2.0f);

	m_d2dContext->SetTransform(translation * m_orientationTransform2D);

    switch(m_renderingOption)
    {
    case RenderingOption::Draw:
        m_d2dContext->DrawGeometry(m_geometryGroup.Get(), m_blackBrush.Get());
        break;

    case RenderingOption::DrawWideStroke:
        m_d2dContext->DrawGeometry(m_geometryGroup.Get(), m_blackBrush.Get(), 20);
        break;

    case RenderingOption::DrawRoundedWideStroke:
        m_d2dContext->DrawGeometry(m_geometryGroup.Get(), m_blackBrush.Get(), 20, m_roundedStrokeStyle.Get());
        break;

    case RenderingOption::DrawDottedWideStroke:
        m_d2dContext->DrawGeometry(m_geometryGroup.Get(), m_blackBrush.Get(), 10, m_dottedStrokeStyle.Get());
        break;

    case RenderingOption::AnimatedDotOffset:
        m_d2dContext->DrawGeometry(m_geometryGroup.Get(), m_blackBrush.Get(), 10, m_animatedStrokeStyle.Get());
        break;

    case RenderingOption::Fill:
        m_d2dContext->FillGeometry(m_geometryGroup.Get(), m_redBrush.Get());
        break;

    case RenderingOption::DrawThenFill:
        m_d2dContext->DrawGeometry(m_geometryGroup.Get(), m_blackBrush.Get(), 10);
        m_d2dContext->FillGeometry(m_geometryGroup.Get(), m_redBrush.Get());
        break;

    case RenderingOption::FillThenDraw:
        m_d2dContext->FillGeometry(m_geometryGroup.Get(), m_redBrush.Get());
        m_d2dContext->DrawGeometry(m_geometryGroup.Get(), m_blackBrush.Get(), 10);
        break;

    case RenderingOption::DrawSimplified:
        m_d2dContext->DrawGeometry(m_simplifiedGeometry.Get(), m_blackBrush.Get(), 10);
        break;

    case RenderingOption::DrawGrosslySimplified:
        m_d2dContext->DrawGeometry(m_grosslySimplifiedGeometry.Get(), m_blackBrush.Get(), 10);
        break;

    case RenderingOption::DrawOutlined:
        m_d2dContext->DrawGeometry(m_outlinedGeometry.Get(), m_blackBrush.Get(), 10);
        break;

    case RenderingOption::FillAndDrawOutlined:
        m_d2dContext->FillGeometry(m_outlinedGeometry.Get(), m_redBrush.Get());
        m_d2dContext->DrawGeometry(m_outlinedGeometry.Get(), m_blackBrush.Get(), 10);
        break;

    case RenderingOption::DrawWidened:
        m_d2dContext->DrawGeometry(m_widenedGeometry.Get(), m_blackBrush.Get(), 1);
        break;

    case RenderingOption::FillWidened:
        m_d2dContext->FillGeometry(m_widenedGeometry.Get(), m_redBrush.Get());
        break;

    case RenderingOption::DrawOutlinedWidened:
        m_d2dContext->DrawGeometry(m_outlinedWidenedGeometry.Get(), m_blackBrush.Get(), 1);
        break;

    case RenderingOption::FillAndDrawOutlinedWidened:
        m_d2dContext->FillGeometry(m_outlinedWidenedGeometry.Get(), m_redBrush.Get());
        m_d2dContext->DrawGeometry(m_outlinedWidenedGeometry.Get(), m_blackBrush.Get(), 10);
        break;
    }

	// Ignore D2DERR_RECREATE_TARGET. This error indicates that the device
	// is lost. It will be handled during the next call to Present.
	HRESULT hr = m_d2dContext->EndDraw();
	if (hr != D2DERR_RECREATE_TARGET)
	{
		DX::ThrowIfFailed(hr);
	}
}

void GeometryVarietiesRenderer::SetFillMode(int fillMode)
{
    m_fillMode = (D2D1_FILL_MODE)fillMode;
    CreateGeometryGroup();
}

void GeometryVarietiesRenderer::SetRenderingOption(GeometryExperimentation::RenderingOption renderingOption)
{
    m_renderingOption = renderingOption;
}

